package cn.com.qmth.stmms.admin.exam;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

import net.sf.json.JSONArray;
import net.sf.json.JSONObject;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;

import cn.com.qmth.stmms.biz.exam.model.ExamStudent;
import cn.com.qmth.stmms.biz.exam.model.ExamSubject;
import cn.com.qmth.stmms.biz.exam.model.MarkGroup;
import cn.com.qmth.stmms.biz.exam.query.ExamStudentSearchQuery;
import cn.com.qmth.stmms.biz.exam.query.SubjectiveScoreSearchQuery;
import cn.com.qmth.stmms.biz.exam.service.ExamQuestionService;
import cn.com.qmth.stmms.biz.exam.service.ExamStudentService;
import cn.com.qmth.stmms.biz.exam.service.MarkGroupService;
import cn.com.qmth.stmms.biz.exam.service.MarkerService;
import cn.com.qmth.stmms.biz.exam.service.SubjectiveScoreService;
import cn.com.qmth.stmms.biz.file.service.FileService;
import cn.com.qmth.stmms.biz.lock.LockService;
import cn.com.qmth.stmms.biz.mark.service.MarkLibraryService;
import cn.com.qmth.stmms.biz.mark.service.MarkSpecialTagService;
import cn.com.qmth.stmms.biz.mark.service.MarkTrackService;
import cn.com.qmth.stmms.biz.user.service.UserService;
import cn.com.qmth.stmms.biz.utils.ScoreItem;
import cn.com.qmth.stmms.common.annotation.Logging;
import cn.com.qmth.stmms.common.annotation.RoleRequire;
import cn.com.qmth.stmms.common.domain.WebUser;
import cn.com.qmth.stmms.common.enums.LockType;
import cn.com.qmth.stmms.common.enums.LogType;
import cn.com.qmth.stmms.common.enums.Role;
import cn.com.qmth.stmms.common.enums.SubjectiveStatus;
import cn.com.qmth.stmms.common.utils.RequestUtils;

@Controller("inspectedController")
@RequestMapping("/admin/exam/inspected")
public class InspectedController extends BaseExamController {

    protected static Logger log = LoggerFactory.getLogger(InspectedController.class);

    @Autowired
    private MarkLibraryService libraryService;

    @Autowired
    private MarkerService markerService;

    @Autowired
    private ExamStudentService studentService;

    @Autowired
    private ExamQuestionService questionService;

    @Autowired
    private MarkTrackService markTrackService;

    @Autowired
    private MarkSpecialTagService markSpecialTagService;

    @Autowired
    private MarkGroupService groupService;

    @Autowired
    private UserService userService;

    @Autowired
    private FileService fileService;

    @Autowired
    private LockService lockService;

    @Autowired
    private SubjectiveScoreService scoreService;

    // 并发处理互斥锁
    private Map<Integer, Integer> currentTaskMap = new HashMap<Integer, Integer>();

    @RequestMapping
    public String list(Model model, HttpServletRequest request, ExamStudentSearchQuery query,
            @RequestParam(required = false) Integer groupNumber,
            @RequestParam(required = false) Double groupStartScore,
            @RequestParam(required = false) Double groupEndScore, @RequestParam(required = false) Double questionScore) {
        int examId = getSessionExamId(request);
        WebUser wu = RequestUtils.getWebUser(request);
        query.setExamId(examId);
        query.setUpload(true);
        query.setAbsent(false);
        query.setBreach(false);
        List<ExamSubject> subjectList = getExamSubject(examId, wu);
        if (query.getSubjectCode() == null && !subjectList.isEmpty()) {
            query.setSubjectCode(subjectList.get(0).getCode());
        }
        if (groupNumber != null || groupStartScore != null || groupEndScore != null || questionScore != null) {
            SubjectiveScoreSearchQuery query2 = new SubjectiveScoreSearchQuery();
            query2.setExamId(examId);
            query2.setSubjectCode(query.getSubjectCode());
            query2.setGroupNumber(groupNumber);
            query2.setGroupStartScore(groupStartScore);
            query2.setGroupEndScore(groupEndScore);
            query2.setScore(questionScore);
            query2.setPageSize(Integer.MAX_VALUE);
            List<Integer> studentIds = scoreService.findStudentIdByQuery(query2);
            query.setStudentIds(studentIds);
        }
        query = studentService.findByQuery(query);
        List<MarkGroup> groupList = groupService.findByExamAndSubject(examId, query.getSubjectCode());
        for (MarkGroup group : groupList) {
            group.setQuestionList(questionService.findByExamAndSubjectAndObjectiveAndGroupNumber(examId,
                    group.getSubjectCode(), false, group.getNumber()));
        }
        model.addAttribute("query", query);
        model.addAttribute("groupList", groupList);
        model.addAttribute("groupNumber", groupNumber);
        model.addAttribute("groupStartScore", groupStartScore);
        model.addAttribute("groupEndScore", groupEndScore);
        model.addAttribute("questionScore", questionScore);
        model.addAttribute("subjectList", subjectList);
        model.addAttribute("statusList", SubjectiveStatus.getOptionList());
        return "modules/exam/inspectedList";
    }

    @Logging(menu = "开始复核", type = LogType.QUERY)
    @RequestMapping("/start")
    @ResponseBody
    public ModelAndView start(HttpServletRequest request, RedirectAttributes redirectAttributes,
            ExamStudentSearchQuery query, @RequestParam(required = false) Integer groupNumber,
            @RequestParam(required = false) Double groupStartScore,
            @RequestParam(required = false) Double groupEndScore, @RequestParam(required = false) Double questionScore) {
        int examId = getSessionExamId(request);
        WebUser wu = RequestUtils.getWebUser(request);
        releaseByUser(wu.getUser().getId());
        ModelAndView view = new ModelAndView("modules/exam/inspected");
        query.setExamId(examId);
        query.setUpload(true);
        query.setAbsent(false);
        query.setBreach(false);
        List<ExamSubject> subjectList = getExamSubject(examId, wu);
        if (query.getSubjectCode() == null && !subjectList.isEmpty()) {
            query.setSubjectCode(subjectList.get(0).getCode());
        }
        if (groupNumber != null || groupStartScore != null || groupEndScore != null || questionScore != null) {
            SubjectiveScoreSearchQuery query2 = new SubjectiveScoreSearchQuery();
            query2.setExamId(examId);
            query2.setSubjectCode(query.getSubjectCode());
            query2.setGroupNumber(groupNumber);
            query2.setGroupStartScore(groupStartScore);
            query2.setGroupEndScore(groupEndScore);
            query2.setScore(questionScore);
            query2.setPageSize(Integer.MAX_VALUE);
            List<Integer> studentIds = scoreService.findStudentIdByQuery(query2);
            query.setStudentIds(studentIds);
        }
        query = studentService.findByQuery(query);
        List<Integer> ids = new ArrayList<Integer>();
        for (ExamStudent student : query.getResult()) {
            ids.add(student.getId());
        }
        view.addObject("fileServer", fileService.getFileServer());
        view.addObject("ids", StringUtils.join(ids, ","));
        return view;
    }

    @RequestMapping("/info")
    @ResponseBody
    public Object info(HttpServletRequest request, @RequestParam Integer studentId) {
        WebUser wu = RequestUtils.getWebUser(request);
        ExamStudent student = studentService.findById(studentId);
        JSONObject result = new JSONObject();
        if (setCurrent(student.getId(), wu.getUser().getId())) {
            result.accumulate("id", student.getId());
            result.accumulate("subjectCode", student.getSubjectCode());
            result.accumulate("subjectName", student.getSubjectName());
            JSONArray array = new JSONArray();
            List<ScoreItem> scores = student.getScoreList(false);
            for (ScoreItem scoreItem : scores) {
                JSONObject obj = new JSONObject();
                obj.accumulate("questionNumber", scoreItem.getMainNumber() + "-" + scoreItem.getSubNumber());
                obj.accumulate("score", scoreItem.getScore());
                array.add(obj);
            }
            result.accumulate("questions", array);

            // MarkGroup group = groupService.findOne(student.getExamId(),
            // student.getSubjectCode(),
            // library.getGroupNumber());
            // List<String> picUrls =
            // fileService.getSliceUris(student.getExamId(),
            // student.getSecretNumber(), 1,
            // student.getSliceCount());
            // List<MarkTrack> markTracks =
            // markTrackService.findByLibraryId(library.getId());
            // List<MarkSpecialTag> markSpecialTagList =
            // markSpecialTagService.findByLibraryId(library.getId());
            // result.accumulate("picUrls", picUrls);
            // result.accumulate("pictureConfig", group.getPictureConfigList());
            // result.accumulate("markTracks", markTracks);
            // result.accumulate("markSpecialTagList", markSpecialTagList);
            result.accumulate("success", true);
        } else {
            result.accumulate("success", false);
        }
        return result;
    }

    @Logging(menu = "评卷复核", type = LogType.UPDATE)
    @RequestMapping("/save")
    @ResponseBody
    public Object save(HttpServletRequest request, @RequestParam Integer studentId) {
        ExamStudent student = studentService.findById(studentId);
        if (student != null && student.getSubjectiveStatus().equals(SubjectiveStatus.MARKED)) {
            studentService.updateSubjectiveStatusAndTime(studentId, SubjectiveStatus.INSPECTED, new Date());
            return true;
        } else {
            return false;
        }
    }

    @RequestMapping("/clear")
    @ResponseBody
    @RoleRequire({ Role.SCHOOL_ADMIN, Role.SUBJECT_HEADER })
    public Object clear(HttpServletRequest request, @RequestParam(required = false) Integer libraryId) {
        WebUser wu = RequestUtils.getWebUser(request);
        if (libraryId != null) {
            releaseTask(libraryId);
        } else {
            releaseByUser(wu.getUser().getId());
        }
        return true;
    }

    @Logging(menu = "取消复核", type = LogType.UPDATE)
    @RequestMapping(value = "/back", method = RequestMethod.POST)
    @ResponseBody
    @RoleRequire({ Role.SCHOOL_ADMIN, Role.SUBJECT_HEADER })
    public JSONObject back(HttpServletRequest request, @RequestParam Integer id) {
        JSONObject obj = new JSONObject();
        ExamStudent student = studentService.findById(id);
        WebUser wu = RequestUtils.getWebUser(request);
        if (student != null && student.getSubjectiveStatus().equals(SubjectiveStatus.INSPECTED)) {
            if (subjectCheck(student.getSubjectCode(), wu)) {
                try {
                    if (lockService.trylock(LockType.STUDENT, student.getId())) {
                        studentService.updateSubjectiveStatusAndTime(student.getId(), SubjectiveStatus.MARKED, null);
                        obj.accumulate("success", true);
                    } else {
                        obj.accumulate("success", false);
                        obj.accumulate("message", "无法取消复核");
                    }
                } catch (Exception e) {
                    obj.accumulate("success", false);
                    obj.accumulate("message", "取消复核失败");
                    log.error("back inspected error", e);
                } finally {
                    lockService.unlock(LockType.STUDENT, student.getId());
                }
            } else {
                obj.accumulate("success", false);
                obj.accumulate("message", "没有操作该评卷任务的权限");
            }
        } else {
            obj.accumulate("success", false);
            obj.accumulate("message", "无法取消复核");
        }
        return obj;
    }

    private boolean setCurrent(Integer taskId, Integer userId) {
        Integer value = currentTaskMap.get(taskId);
        if (value == null) {
            synchronized (currentTaskMap) {
                value = currentTaskMap.get(taskId);
                if (value == null) {
                    currentTaskMap.put(taskId, userId);
                    return true;
                } else {
                    return false;
                }
            }
        } else {
            return false;
        }
    }

    private void releaseTask(Integer taskId) {
        synchronized (currentTaskMap) {
            currentTaskMap.remove(taskId);
        }
    }

    private void releaseByUser(Integer userId) {
        Set<Integer> taskIds = new HashSet<>();
        taskIds.addAll(currentTaskMap.keySet());
        synchronized (currentTaskMap) {
            for (Integer taskId : taskIds) {
                Integer value = currentTaskMap.get(taskId);
                if (value != null && value.equals(userId)) {
                    currentTaskMap.remove(taskId);
                }
            }
        }
    }
}
